'' ------------------------------------------------------------------
'' SMACKRO v0.6
'' ------------------------------------------------------------------
''  Copyright (C) 2007-2008, Jeff Marshall (coder@execulink.com)
''
''  This library is free software; you can redistribute it and/or
''  modify it under the terms of the GNU Lesser General Public
''  License as published by the Free Software Foundation; either
''  version 2.1 of the License, or (at your option) any later version.
''
''  This library is distributed in the hope that it will be useful,
''  but WITHOUT ANY WARRANTY; without even the implied warranty of
''  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
''  Lesser General Public License for more details.
''
''  You should have received a copy of the GNU Lesser General Public
''  License along with this library; if not, write to the Free Software
''  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
''  MA  02110-1301  USA
'' ------------------------------------------------------------------

#include once "smackro.bi"

'' ------------------------------------------------------------------
'' Private Stuff

const OPT_SPACE_IS_SEPERATOR = 1
const OPT_COMMA_IS_SEPERATOR = 2

'' holds the name and value for one macro variable
type MacroVarT
	n as string		'' name
	v as string		'' value
	noclear as integer
end type

'' holds the name and function pointer for a macro function
type MacroFuncT
	n as string			'' name
	minargs as integer	'' minimum number of arguments ( not including function name )
	func as function( byval nargs as integer, args() as string ) as string
end type

declare function __MacroEvalFunc _
	( _
		 byref text as string _
	) as string

declare function __MacroGetExpr _
	( _
		 byref text as string, _
		 byval start as integer, _
		 byval opts as integer _
	) as integer

'' array to hold all macro variables 
dim shared vars() as MacroVarT, nVars as integer
dim shared funcs() as MacroFuncT, nFuncs as integer

'' ------------------------------------------------------------------
'' Compatibility Stuff

#ifdef __FB_LANG__
	#if __FB_LANG__ = "qb"
		#define PUBLIC_
		#define PRIVATE_
	#else
		#define PUBLIC_ public
		#define PRIVATE_ private
	#endif
#endif

'' ------------------------------------------------------------------
'' Set a macro variable's value.  If the variable does
'' not already exist, it is created.  Names are case
'' insensitive.

PUBLIC_ sub MacroSetVar _
	( _
		 byref n as string, _
		 byref v as string, _
		 byval noclear as integer _
	)

	dim i as integer
	for i = 0 to nVars - 1
		if( ucase(n) = vars(i).n ) then
			exit for
		end if
	next
	if( i >= nVars ) then
		nVars += 1
		if( nVars = 1 ) then
			redim vars(0 to nVars - 1)
		else
			redim preserve vars( 0 to nVars - 1 )
		end if
		i = nVars - 1
	end if
		
	vars(i).n = ucase(n)
	vars(i).v = v
	vars(i).noclear = noclear

end sub

'' ------------------------------------------------------------------
'' Get a macro variable's value by name.  If the name
'' does not exist, returns an empty string

PUBLIC_ function MacroGetVar _
	( _
		 byref n as string _
	) as string

	dim i as integer
	function = ""
	for i = 0 to nVars - 1
		if( ucase(n) = vars(i).n ) then
			function = vars(i).v
			exit for
		end if
	next
end function

'' ------------------------------------------------------------------
'' Bind a freebasic function to a macro name.
'' The function called must have the signature:
'' function func_name( byval nargs as integer, args() as string ) as string

PUBLIC_ sub MacroBind _
	( _
		byref n as string, _
		byval func as function( byval nargs as integer, args() as string ) as string, _
		byval minargs as integer _
	)
	
	dim i as integer

	for i = 0 to nFuncs - 1
		if( ucase(n) = funcs(i).n ) then
			exit for
		end if
	next

	if( i >= nFuncs ) then
		nFuncs += 1
		if( nFuncs = 1 ) then
			redim funcs(0 to nFuncs - 1)
		else
			redim preserve funcs( 0 to nFuncs - 1 )
		end if
		i = nFuncs - 1
	end if

	funcs(i).n = ucase(n)
	funcs(i).func = func
	funcs(i).minargs = minargs

end sub

'' ------------------------------------------------------------------
'' Execute a macro.  nargs is the number of arguments. args() is a
'' zero based array of arguments.  args(0) is the name of the function
'' to execute.  Returns the string result of the function.

PUBLIC_ function MacroExec _
	( _
		 byval nargs as integer, _
		 args() as string _
	) as string

	dim i as integer

	function = ""

	if( nargs = 0 ) then
		exit function
	end if

	for i = 0 to nfuncs - 1
		if( funcs(i).n = ucase(args(0)) ) then
			exit for
		end if
	next

	if( i >= nfuncs ) then
		function = MacroEval( MacroGetVar( MacroEval( args(0) ) ) )
		exit function
	end if

	if( nargs > funcs(i).minargs ) then
		function = funcs(i).func( nargs, args() )

	else
		function = ""
		
	end if

end function

'' ------------------------------------------------------------------
'' Get an expression sub-string.  Everything up to the next closing
'' paren.  Returns the index of the closing paren, or one past the
'' length of text if not found.

PRIVATE_ function __MacroGetExpr _
	( _
		 byref text as string, _
		 byval start as integer, _
		 byval opts as integer _
	) as integer

	dim as integer i, n, level

	i = start
	n = len( text )
	level = 1

	while( i <= n )
		select case mid(text,i,1)
		case "$"
			i += 1
			if( i <= n ) then
				if( mid(text,i,1) = "(" ) then
					level += 1
				end if
			end if

		case ")"
			level -= 1
			if( level = 0 ) then
				exit while
			end if
		case ","
			if( (opts and OPT_COMMA_IS_SEPERATOR) <> 0 ) then
				if( level <= 1 ) then
					exit while
				end if
			end if

		case " "
			if( (opts and OPT_SPACE_IS_SEPERATOR) <> 0 ) then
				if( level <= 1 ) then
					exit while
				end if
			end if

		end select

		if( i <= n ) then
			i += 1
		end if
	wend

	function = i

end function


'' ------------------------------------------------------------------
'' Evaluates an expression in the form "function argument"
'' if the function is not found, it is evaluated as a variable
'' and the argument is ignored.

PRIVATE_ function __MacroEvalFunc _
	( _
		 byref text as string _
	) as string

	dim as integer i, j, n, opts, nargs
	redim args(0) as string
	
	nargs = 0
	n = len(text)
	opts = OPT_SPACE_IS_SEPERATOR or OPT_COMMA_IS_SEPERATOR

	'' Get the args
	i = 1
	while( i <= n )
		j = __MacroGetExpr( text, i, opts )
		nargs += 1
		if( nargs = 1 ) then
			redim args(0 to nargs - 1)
		else
			redim preserve args( 0 to nargs - 1 )
		end if
		args( nargs - 1 ) = mid( text, i, j - i )
		i = j + 1
		opts and= NOT OPT_SPACE_IS_SEPERATOR
	wend

	function = MacroExec( nargs, args() )

end function

'' ------------------------------------------------------------------
'' Evaluates a text string by expanding all variables and evaluating
'' functions.

PUBLIC_ function MacroEval _
	( _
		 byref text as string _
	) as string
	
	dim ret as string
	dim as integer i,j,n

	ret = ""
	i = 1
	n = len(text)

	while( i <= n )
		if( mid(text,i,1) = "$" ) then 
			i += 1
			if( i <= len(text) ) then
				if( mid(text,i,1) = "(" ) then
					i += 1
					j = __MacroGetExpr( text, i, 0 )
					ret += __MacroEvalFunc( mid( text, i, j - i ))
					i = j + 1
				else
					ret += mid(text,i,1)
					i += 1
				end if
			end if
		else
			ret += mid(text,i,1)
			i += 1
		end if
	wend

	function = ret

end function

'' ------------------------------------------------------------------
'' INTRINSIC FUNCTIONS
'' ------------------------------------------------------------------

'' $(ucase text)
DEFINE_MACRO_FUNC( ucase )
	function = ucase( MacroEval( args(1) ) )
END_MACRO_FUNC

'' $(lcase text)
DEFINE_MACRO_FUNC( lcase )
	function = lcase( MacroEval( args(1) ) )
END_MACRO_FUNC

'' $(left test,n)
DEFINE_MACRO_FUNC( left )
	function = left( MacroEval( args(1) ), val( MacroEval( args(2) ) ) )
END_MACRO_FUNC

'' $(right test,n)
DEFINE_MACRO_FUNC( right )
	function = right( MacroEval( args(1) ), val( MacroEval( args(2) ) ) )
END_MACRO_FUNC

'' $(concat a [, ...] )
DEFINE_MACRO_FUNC( concat )
	dim i as integer, ret as string
	for i = 1 to nargs - 1
		ret += MacroEval( args(i) )
	next i
	function = ret
END_MACRO_FUNC

'' ------------------------------------------------------------------
'' Initialize this module.  It doesn't need to be called, but
'' intrinsic functions will only be available if it is.

PUBLIC_ sub MacroInit _
	( _
	)

	'' function, minargs
	BIND_MACRO_FUNC( ucase, 1 )
	BIND_MACRO_FUNC( lcase, 1 )
	BIND_MACRO_FUNC( left, 2 )
	BIND_MACRO_FUNC( right, 2 )
	BIND_MACRO_FUNC( concat, 1 )
	
end sub

'' ------------------------------------------------------------------
'' Clear all variables, but do not clear any bound functions.

PUBLIC_ sub MacroClearVars _
	( _
	)
	
	dim i as integer, j as integer
	
	j = 0
	
	for i = 0 to nVars - 1
		if( vars(i).noclear <> 0 ) then
			if( i <> j ) then
				vars(j) = vars(i)
			end if
			j += 1
		end if
	next
	nVars = j

	if( nVars > 0 ) then
		redim preserve vars(0 to nVars - 1)
	else
		redim vars(0)
	end if

end sub

'' ------------------------------------------------------------------
'' Clear all functions.

PUBLIC_ sub MacroClearFuncs _
	( _
	)
	
	nFuncs = 0
	redim funcs(0)

end sub
